package webwisdom.tango.object;

import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import webwisdom.tango.messages.*;
import webwisdom.tango.consts.*;
import webwisdom.tango.*;

/**
  * The central class of the TangoObject interface. It manages distributed objects and
  * the flow of data between corresponding objects on different clients.
  * 
  * @author Piotr Sokolowski
  * @version $Revision: 1.5 $
  * @see webwisdom.tango.object.TObject
  * @see webwisdom.tango.object.MessageOutputStream
  */
public class TNode extends TAgent implements Runnable,TDataListener
{
  private Thread th;
  private Vector msgQueue;

  private final static byte TOBJECT_CRE=0;
  private final static byte TOBJECT_DEST=1;
  private final static byte TOBJECT_DAT=2;
  private final static byte NON_TOBJECT_DAT=3;
  private final static byte INIT_REQ=4;
  private final static byte INIT_RES=5;
  private final static byte LAST_CODE=5;

  private Hashtable tobject;
  private Hashtable trobject;
  private boolean initialized;
  
  private AppEventMessage aem;
  private boolean running;
  private Object  thsync;
  private ClassLoader cload;
  private long lastSend;
  private static int k=0; // variable for key generation.

  /**
    * Creates a new TNode object.
    * @param tlagent An object implementing TLAgent interface for communication with Tango System
    * @param cloader Class loader that can load classes that may be remotely created. Usually it is the class loader of the Applet.
    * <PRE>
    * 	new TNode(tlagent,this.getClass().getClassLoader());
    * </PRE>
    */
  public TNode(TLAgent tlagent,ClassLoader cloader)
    {
      super(tlagent);
      cload=cloader;
      initialized=false;
      thsync=new Object();
      tobject=new Hashtable();
      trobject=new Hashtable();
      running=false;
      msgQueue=new Vector();
      addTDataListener(this);
    }

  /**
    * Requests initialization information. The starting client is synchronized with the state of the session.
    */
  public void init()
    {
      if(isMaster())
	initialized=true;
      else
	try
	  {
	    MessageOutputStream mos=new MessageOutputStream(this,false,new ByteArrayOutputStream());
	    mos.writeBoolean(false);
	    mos.writeByte(INIT_REQ);
	    mos.send();
	  }
      catch(Exception e)
	{
	  e.printStackTrace();
	}
    }

  /**
    * Does not belong to API
    * services the internal message queue.
    */
  protected final byte[] messageNext()
    {
      byte b[];
      synchronized(msgQueue)
	{
	  try
	    {
	      while(msgQueue.size()<=0)
		msgQueue.wait();
	    }
	  catch(Exception e)
	    {
	      e.printStackTrace();
	    }
	  b=(byte[])msgQueue.elementAt(0);
	  msgQueue.removeElementAt(0);
	  return b;
	}
    }

  /**
    * Does not belong to API
    * needed for creating new Thread.
    */
  public final void run()
    {
      while(true)
	{
	  receiveAnyData();
	  synchronized(thsync)
	    {
	      if(running==false)
		{
		  th=null;
		  return;
		}
	    }
	}
    }

  /**
    * Does not belong to API
    * receives incoming messages from TangoSystem.
    * puts messages to the queue.
    */
  public final void receive(byte b[])
    {
      synchronized(msgQueue)
	{
	  msgQueue.addElement(b);
	  msgQueue.notify();
	}
    }

  /**
    * Creates a new MessageOutputStream associated with an object specified in the constructor.
    * The Message when send is automaticly routed to the corresponding objects on other clients.
    * @param tobject TObject which requests the new MessageOutputStream. The data written to the MessageOutputStream will then be send to all corresponding TObjects. If null or not TObject the message will be passed to receiveData method of TNode-s.
    * @param isRemovable specifies whether the message may be discarded in case of congestion.
    * @return A new MessageOutputStream
    */
  public MessageOutputStream getMOS(Object tobject,boolean isRemovable)
    {
      try
	{
	  MessageOutputStream mos=new MessageOutputStream(this,isRemovable,new ByteArrayOutputStream());
	  mos.writeBoolean(isRemovable);
	  if(tobject instanceof TObject)
	    {
	      mos.writeByte(TOBJECT_DAT);
	      mos.writeUTF((String)trobject.get(tobject));
	    }
	  else
	    {
	      mos.writeByte(NON_TOBJECT_DAT);
	    }
	  return mos;
	}
      catch(Exception e)
	{
	  e.printStackTrace();
	}
      return null;
    }

  /**
    * Not part of the API
    * used by the MessageOutputStream to send itself
    */
  final void sendData(MessageOutputStream mos,boolean isRemovable)
    {
      if(isRemovable==false || System.currentTimeMillis()-lastSend>500)
	{
	  send(mos.getByteStream().toByteArray());
	  lastSend=System.currentTimeMillis();
	}
    }


  /** Not part of the API
    * takes one mwssage from the queue and parses it.
    */
  private final void receiveAnyData()
    {
      try
	{
	  ByteArrayInputStream bis=new ByteArrayInputStream(messageNext());
	  DataInputStream dis=new DataInputStream(bis);
	  boolean isRemovable=dis.readBoolean();
	  if(isRemovable && msgQueue.size()>1)
	    return;
	  byte b=dis.readByte();
	  switch(b)
	    {
	    case TOBJECT_CRE:
	      {
		String is=dis.readUTF();
		String oname=dis.readUTF();
		//Class c=Class.forName(oname);
		Class c=cload.loadClass(oname);
		TObject oto=(TObject)tobject.get(is);
		if(oto==null)
		  {
		    Class cls[]={this.getClass(),Class.forName("java.io.DataInputStream")};
		    Constructor cn=c.getConstructor(cls);
		    Object tb[]={this,dis};
		    TObject to=(TObject)cn.newInstance(tb);
		    tobject.put(is,to);
		    trobject.put(to,is);
		  }
		else
		  oto.update(dis);
		break;
	      }
	    case TOBJECT_DEST:
	      {
		String is=dis.readUTF();
		TObject to=(TObject)tobject.get(is);
		tobject.remove(is);
		to.dispose();
		break;
	      }
	    case TOBJECT_DAT:
	      {
		String is=dis.readUTF();
		TObject to=(TObject)tobject.get(is);
		to.receive(dis);
		break;
	      }
	    case NON_TOBJECT_DAT:
	      {
		receiveData(dis);
		break;
	      }
	    case INIT_REQ:
	      {
		if(isMaster())
		  {
		    MessageOutputStream mos=new MessageOutputStream(this,false,new ByteArrayOutputStream());
		    mos.writeBoolean(false);
		    mos.writeByte(INIT_RES);
		
		    ByteArrayOutputStream bos;
		    DataOutputStream dos;

		    // sending general non-object specific initialization info
		    bos=new ByteArrayOutputStream();
		    dos=new DataOutputStream(bos);
		    writeInit(dos);
		    mos.writeInt(bos.size());
		    bos.writeTo(mos);

		    // sending info needed to create all objects
		    Enumeration e=tobject.keys();
		    while(e.hasMoreElements())
		      {
			String key=(String)e.nextElement();
			TObject to=(TObject)tobject.get(key);
			bos=new ByteArrayOutputStream();
			dos=new DataOutputStream(bos);
			dos.writeUTF(key);
			dos.writeUTF(to.getClass().getName());
			to.writeTo(dos);
			mos.writeInt(bos.size());
			bos.writeTo(mos);
		      }
		    mos.send();
		  }
		break;
	      }
	    case INIT_RES:
	      {
		if(initialized==false)
		  {
		    int len;
		    byte barray[];
		    DataInputStream ldis;

		    // reading general initialization info
		    len=dis.readInt();
		    barray=new byte[len];
		    dis.readFully(barray);
		    ldis=new DataInputStream(new ByteArrayInputStream(barray));
		    readInit(ldis);
	    
		    while(dis.available()>0)
		      {
			len=dis.readInt();
			barray=new byte[len];
			dis.readFully(barray);
			ldis=new DataInputStream(new ByteArrayInputStream(barray));
			String key=ldis.readUTF();
			String cname=ldis.readUTF();

			//Class  c=cload.forName(cname);
			Class  c=cload.loadClass(cname);
			TObject to;
		    
			to=(TObject)tobject.get(key);
			if(to==null)
			  {
			    Class cls[]={this.getClass(),Class.forName("java.io.DataInputStream")};
			    Constructor cns[]=c.getConstructors();
			    Constructor cn=c.getConstructor(cls);
			    Object tb[]={this,ldis};
			    to=(TObject)cn.newInstance(tb);
			    tobject.put(key,to);
			    trobject.put(to,key);
			  }
			else
			  {
			    if(to.getClass().equals(c))
			      to.update(ldis);
			  }
		      }
		    initialized=true;
		    synchronized(this)
		      {
			notifyAll();
		      }
		  }
		break;
	      }
	    default:
	      throw new Exception("Unable to interpret the data");
	    }
	}
      catch(Exception e)
	{
	  e.printStackTrace();
	}
    }

  /**
    * Called if received non TObject data
    * @param dis data passed from another application
    * @exception Exception (will be ignored)
    */
  protected void receiveData(DataInputStream dis) throws Exception
    {
    }

  /**
    * Called when sending state to another application. This method is called before sending all TObjects.
    * @param dos DataOutputStream to which all state data is to be written
    * @exception Exception (will be ignored)
    */
  protected void writeInit(DataOutputStream dos) throws Exception
    {
    }

  /**
    * called upon receiving of state of the session. This is called before creation of
    * TObjects.
    * @param dis initialization data (the same as written by writeInit)
    * @exception Exception (will be ignored)
    */
  protected void readInit(DataInputStream dis) throws Exception
    {
    }

  /**
    * adds a new TObject with automaticaly generated key, corresponding TObject-s will be created
    * on all clients in the session
    * @param to TObject to be distributed.
    */
  public final void add(TObject to)
    {
      try
	{
	  String id=newId();
	  tobject.put(id,to);
	  trobject.put(to,id);
	  MessageOutputStream mos=new MessageOutputStream(this,false,new ByteArrayOutputStream());
	  mos.writeBoolean(false);
	  mos.writeByte(TOBJECT_CRE);
	  mos.writeUTF((String)trobject.get(to));
	  mos.writeUTF(to.getClass().getName());
	  to.writeTo(mos);
	  mos.send();
	}
      catch(Exception e)
	{
	  e.printStackTrace();
	}
    }
  
  /**
    * adds a new TObject with a specified key. The corresponding objects will be created only if 
    * objects with the same key do not already exist on other clients.
    * @param to TObject to be distributed
    * @param key key which will be assigned to the TObject to
    */
  public final void add(String key,TObject to)
    {
      try
	{
	  TObject oto=(TObject)tobject.get(key);
	  if(oto!=null)
	    {
	      if(oto!=to)
		{
		  remove(oto);
		  oto.dispose();
		}
	      else
		return;
	    }
	  tobject.put(key,to);
	  trobject.put(to,key);
	  if(initialized)
	    {
	      MessageOutputStream mos=new MessageOutputStream(this,false,new ByteArrayOutputStream());
	      mos.writeBoolean(false);
	      mos.writeByte(TOBJECT_CRE);
	      mos.writeUTF((String)trobject.get(to));
	      mos.writeUTF(to.getClass().getName());
	      to.writeTo(mos);
	      mos.send();
	    }
	}
      catch(Exception e)
	{
	}
    }

  /**
    * removes a TObject. The corresponding TObjects are also destroyed on all other clients.
    * @param to TObject to be removed.
    */
  public final void remove(TObject to)
    {
      try
	{
	  Object key=trobject.get(to);
	  tobject.remove(key);
	  trobject.remove(to);
	  MessageOutputStream mos=new MessageOutputStream(this,false,new ByteArrayOutputStream());
	  mos.writeBoolean(false);
	  mos.writeByte(TOBJECT_DEST);
	  mos.writeUTF((String)key);
	  mos.send();
	}
      catch(Exception e)
	{
	  e.printStackTrace();
	}
    }
  
  /**
    * Allows to find a TObject identified by a String.
    * @param id String identifying the TObject. It equals to the key with which the TObject has been added to the TNode.
    * @return TObject that has been added with the id as the key, or null if the key was not found.
    */
  public TObject getTObject(String id)
    {
      return (TObject)tobject.get(id);
    }
  
  /**
    * Allows to find the key of a given TObject.
    * @param to the reference to the TObject
    * @return key corresponding to the given TObject or null if not found.
    */
  public String getKey(TObject to)
    {
      return (String)trobject.get(to);
    }

  /**
    * Creates a new unique id that can be used as a key.
    * the id is guaranteed to be unique for all clients in the session.
    * @return newly generated id.
    */
  public  String newId()
    {
      String s=getUserName()+"\n"+k;
      k++;
      return s;
    }

  /**
    * called to dispose of the TNode.
    * stops threads, etc.
    */
  public void dispose()
    {
      stop();
    }

  /**
    * stops the queue thread
    */
  public final void stop()
    {
      synchronized(thsync)
	{	
	  running=false;
	}
    }

  /**
    * starts the queue thread
    */
  public final void start()
    {
      synchronized(thsync)
	{
	  running=true;
	  if(th==null)
	    {
	      th=new Thread(this);
	      th.start();   
	    }
	}
    }

  /**
    * Allows waiting for TNode to synchronize with the session.
    * The thread will leave this method only after TNode has received full
    * initialization information and is ready to action
    */
  public final void waitForInitialization()
    {
      synchronized(this)
	{
	  if(!initialized)
	    {
	      try
		{
		  wait();
		}
	      catch(Exception e)
		{
		  e.printStackTrace();
		}
	    }
	}
    }
}

