package webwisdom.tango.object;
import java.io.*;

/**
  * TObject interface allows objects to register in TNode object. Such registered TObject is
  * automaticaly replicated on all other clients. The original TObject and replicated ones
  * form a group, which can internally exchange messages.
  *
  * If the TObject should be replicated automaticaly on other clients, it must contain a special
  * constructor:
  * <PRE>
  * 	public ConstructorName(TNode tn,DataInputStream dis);
  * </PRE>
  * The dis parameter will contain data written by writeTo method of the corresponding TObject.
  * 
  * @author Piotr Sokolowski
  * @version $Revision: 1.3 $
  * @see webwisdom.tango.object.TObject
  * @see webwisdom.tango.object.MessageOutputStream
  */
public interface TObject
{
  /**
    * Is called when TNode wants to replicate the TObject. The TObject should write
    * there data which would allow recreating of its state.
    * @param dos The stream to which the data about current state should be written
    * @exception Exception (will be ignored)
    */
  public void writeTo(DataOutputStream dos) throws Exception;

  /**
    * Is called when the object is removed from the system.
    * (usually when the corresponding object on other client was destroyed)
    */
  public void dispose();
  
  /**
    * Is called when a message from a corresponding object arrives.
    * @param dis DataInputStream containing received data.
    * @exception Exception (will be ignored)
    */
  public void receive(DataInputStream dis) throws Exception;
  /**
    * Is called when an updating message is received.
    * This message was generated by the writeTo method of the corresponding
    * TObject.
    * @param dis contains update data.
    * @exception Exception (will be ignored)
    */
  public void update(DataInputStream dis) throws Exception;
}
